package org.drools.integrationtests;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.ObjectInput;
import java.io.ObjectInputStream;
import java.io.ObjectOutput;
import java.io.ObjectOutputStream;


import junit.framework.TestCase;


import org.drools.Cheese;
import org.drools.Person;
import org.drools.RuleBase;
import org.drools.RuleBaseFactory;
import org.drools.SerializeAddress;
import org.drools.StatefulSession;
import org.drools.compiler.PackageBuilder;
import org.drools.rule.Package;

public class DynamicRulesMoreTest extends TestCase {
    
	
	/*
	 *  Here I am inserting data points which are not used by any rule (e.g Person).
	 *  Later adding rule (e.g. Rule: 'match Person') for those data points.
	 *  
	 *  Result: Pkg/Rule addition is failing with ClassCastException
	 */
    public void testSerializeAdd_ClassCastException() throws Exception {

        //Create a rulebase, a session, and test it
        RuleBase ruleBase = RuleBaseFactory.newRuleBase( );
        PackageBuilder builder = new PackageBuilder();
        builder.addPackageFromDrl( new InputStreamReader( getClass().getResourceAsStream( "test_Dynamic1_0.drl" ) ) );
        Package pkg = serialisePackage( builder.getPackage() );
        System.out.println("Adding Pkg: org.drools.test with Rule: 'like stilton'");
        ruleBase.addPackage( pkg );
        
        StatefulSession session = ruleBase.newStatefulSession();
       
        session.insert( new Cheese( "stilton", 10 ) );
        session.insert( new Cheese( "brie", 10 ) );
        session.insert( new Person( "bob", 10 ) );
        
        System.out.println("Firing Rules....");
        session.fireAllRules();
  
        byte[] serializedSession = serializeOut( session );
        byte[] serializedRulebase = serializeOut( ruleBase );
        
        session.dispose();
        System.out.println("2nd Run..");
        
        // now recreate the rulebase, deserialize the session and test it
        ruleBase = (RuleBase) serializeIn( serializedRulebase );
        
        session = ruleBase.newStatefulSession( new ByteArrayInputStream( serializedSession ) );
       
        
        builder = new PackageBuilder();
        builder.addPackageFromDrl( new InputStreamReader( getClass().getResourceAsStream( "test_Dynamic3_0.drl" ) ) );
        pkg = serialisePackage( builder.getPackage() );
        
        System.out.println("Adding Pkg: org.drools.newtest with Rule: 'match Person'");
        ruleBase.addPackage( pkg );
        
        session.insert( new Cheese( "stilton", 20 ) );
        session.insert( new Cheese( "brie", 20 ) );
        session.insert( new Person( "bob", 20 ) );
        System.out.println("Firing Rules....");
        session.fireAllRules();
 
        serializedSession = null;
        serializedRulebase = null;
        
        serializedSession = serializeOut( session );
        serializedRulebase = serializeOut( ruleBase );
        
        session.dispose();
       
    }
    
    
    /*
	 *  Here I am inserting data points(e.g. Cheese) with  'stilton' / 'brie' as type value.
	 *  Initially I had only 1 rule ('like stilton') for data points(e.g. Cheese) having type as 'stilton'.
	 *
	 *  Later added new rule ('like brie')  for data points(e.g. Cheese) having type as 'brie'. 
	 *  
	 *  Result: new rule is not getting fired for new data points having type as 'brie'.
	 *          Only for old data points having type as 'brie' the new rule got fired.
	 */
    
    
    public void testSerializeAdd_newRuleNotFiredForNewData() throws Exception {

        //Create a rulebase, a session, and test it
        RuleBase ruleBase = RuleBaseFactory.newRuleBase( );
        PackageBuilder builder = new PackageBuilder();
        builder.addPackageFromDrl( new InputStreamReader( getClass().getResourceAsStream( "test_Dynamic1_0.drl" ) ) );
        Package pkg = serialisePackage( builder.getPackage() );
        System.out.println("Adding Pkg: org.drools.test with Rule: 'like stilton'");
        ruleBase.addPackage( pkg );
        
        StatefulSession session = ruleBase.newStatefulSession();
       
        session.insert( new Cheese( "stilton", 10 ) );
        session.insert( new Cheese( "brie", 10 ) );
        
        System.out.println("Firing Rules....");
        session.fireAllRules();
  
        byte[] serializedSession = serializeOut( session );
        byte[] serializedRulebase = serializeOut( ruleBase );
        
        session.dispose();

        
        System.out.println("2nd Run..");
        
        // now recreate the rulebase, deserialize the session and test it
        ruleBase = (RuleBase) serializeIn( serializedRulebase );
        
        session = ruleBase.newStatefulSession( new ByteArrayInputStream( serializedSession ) );
       
        
        builder = new PackageBuilder();
        builder.addPackageFromDrl( new InputStreamReader( getClass().getResourceAsStream( "test_Dynamic1_1.drl" ) ) );
        pkg = serialisePackage( builder.getPackage() );
        
        System.out.println("Adding Pkg: org.drools.test with Rule: 'like brie'");
        ruleBase.addPackage( pkg );
        
        session.insert( new Cheese( "stilton", 20 ) );
        
        /*
         * Rule is not getting fired for the bellow data points..
         */
        session.insert( new Cheese( "brie", 20 ) );
        session.insert( new Cheese( "brie", 30 ) );
        
        System.out.println("Firing Rules....");
        session.fireAllRules();
        
        System.out.println("Rule 'like brie' is not fired for newly inserted data points (i.e  new Cheese( 'brie', 20 ) ) &  new Cheese( 'brie', 30 )).");
        
        serializedSession = null;
        serializedRulebase = null;
        
        serializedSession = serializeOut( session );
        serializedRulebase = serializeOut( ruleBase );
        
        session.dispose();  
       
    }
    
  
    /*
     *  Works Fine if both the scenarios mentioned above is skiped. 
     */
    
    public void testSerializeAdd() throws Exception {

        //Create a rulebase, a session, and test it
        RuleBase ruleBase = RuleBaseFactory.newRuleBase( );
        PackageBuilder builder = new PackageBuilder();
        builder.addPackageFromDrl( new InputStreamReader( getClass().getResourceAsStream( "test_Dynamic1_0.drl" ) ) );
        Package pkg = serialisePackage( builder.getPackage() );
        
        System.out.println("Adding Pkg: org.drools.test with Rule: 'like stilton'");
        ruleBase.addPackage( pkg );
        
        StatefulSession session = ruleBase.newStatefulSession();
       
        session.insert( new Cheese( "stilton", 10 ) );
        session.insert( new Cheese( "brie", 10 ) );
        
        System.out.println("Firing Rules....");
        session.fireAllRules();
  
        byte[] serializedSession = serializeOut( session );
        byte[] serializedRulebase = serializeOut( ruleBase );
        
        session.dispose();

        
        System.out.println("2nd Run..");
        
        // now recreate the rulebase, deserialize the session and test it
        ruleBase = (RuleBase) serializeIn( serializedRulebase );
        
        session = ruleBase.newStatefulSession( new ByteArrayInputStream( serializedSession ) );
       
        
        builder = new PackageBuilder();
        builder.addPackageFromDrl( new InputStreamReader( getClass().getResourceAsStream( "test_Dynamic3_0.drl" ) ) );
        pkg = serialisePackage( builder.getPackage() );
        
        System.out.println("Adding Pkg: org.drools.newtest with Rule: 'match Person'");
        ruleBase.addPackage( pkg );
        
        session.insert( new Cheese( "stilton", 20 ) );
        session.insert( new Cheese( "brie", 20 ) );
        session.insert( new Person( "bob", 20 ) );
        session.insert( new Person( "bob", 30 ) );
        System.out.println("Firing Rules....");
        session.fireAllRules();
 
        serializedSession = null;
        serializedRulebase = null;
        
        serializedSession = serializeOut( session );
        serializedRulebase = serializeOut( ruleBase );
        
        session.dispose();
        
        System.out.println("3rd Run..");
        
        // now recreate the rulebase, deserialize the session and test it
        ruleBase = (RuleBase) serializeIn( serializedRulebase );
        
        session = ruleBase.newStatefulSession( new ByteArrayInputStream( serializedSession ) );
       
        
        builder = new PackageBuilder();
        builder.addPackageFromDrl( new InputStreamReader( getClass().getResourceAsStream( "test_Dynamic1_2.drl" ) ) );
        pkg = serialisePackage( builder.getPackage() );
        
        System.out.println("Adding Pkg: org.drools.test with Rule: 'match SerializeAddress'");
        ruleBase.addPackage( pkg );
        
        session.insert( new Cheese( "stilton", 40 ) );
        session.insert( new Cheese( "brie", 40 ) );
        session.insert( new Person( "bob", 40 ) );
        session.insert( new Person( "bob", 40 ) );
        session.insert( new SerializeAddress( "bangalore" ) );
        session.insert( new SerializeAddress( "India" ) );
        
        System.out.println("Firing Rules....");
        session.fireAllRules();
 
        serializedSession = null;
        serializedRulebase = null;
        
        serializedSession = serializeOut( session );
        serializedRulebase = serializeOut( ruleBase );
        
        session.dispose();
        
        System.out.println("4th Run..");
        
        // now recreate the rulebase, deserialize the session and test it
        ruleBase = (RuleBase) serializeIn( serializedRulebase );
        
        session = ruleBase.newStatefulSession( new ByteArrayInputStream( serializedSession ) );
       
        
        session.insert( new Cheese( "stilton", 50 ) );
        session.insert( new Cheese( "brie", 50 ) );
        session.insert( new Person( "bob", 50 ) );
        session.insert( new Person( "bob", 50 ) );
        session.insert( new SerializeAddress( "Tripura" ) );
        session.insert( new SerializeAddress( "Agartala" ) );
        
        System.out.println("Firing Rules....");
        session.fireAllRules();
 
        serializedSession = null;
        serializedRulebase = null;
        
        serializedSession = serializeOut( session );
        serializedRulebase = serializeOut( ruleBase );
        
        session.dispose();
       
    }
    
    /*
     * I have tried both the scenarios
     * 1. Remove a rule from a pkg.
     * 2. Remove a pkg
     * 
     * But both cases after inserting associated data points (i.e data points which are used to fire/activate the removed rule)
     * session.fireAllRules() is throwing NoClassDefFoundError
     * 
     */
    public void testSerializeAddRemove_NoClassDefFoundError() throws Exception {

        //Create a rulebase, a session, and test it
        RuleBase ruleBase = RuleBaseFactory.newRuleBase( );
        PackageBuilder builder = new PackageBuilder();
        builder.addPackageFromDrl( new InputStreamReader( getClass().getResourceAsStream( "test_Dynamic1_0.drl" ) ) );
        Package pkg = serialisePackage( builder.getPackage() );
        
        System.out.println("Adding Pkg: org.drools.test with Rule: 'like stilton'");
        
        ruleBase.addPackage( pkg );
        
        StatefulSession session = ruleBase.newStatefulSession();
       
        session.insert( new Cheese( "stilton", 10 ) );
        session.insert( new Cheese( "brie", 10 ) );
        
        System.out.println("Firing Rules....");
        session.fireAllRules();
  
        byte[] serializedSession = serializeOut( session );
        byte[] serializedRulebase = serializeOut( ruleBase );
        
        session.dispose();

        
        System.out.println("2nd Run..");
        
        // now recreate the rulebase, deserialize the session and test it
        ruleBase = (RuleBase) serializeIn( serializedRulebase );
        
        session = ruleBase.newStatefulSession( new ByteArrayInputStream( serializedSession ) );
       
        
        builder = new PackageBuilder();
        builder.addPackageFromDrl( new InputStreamReader( getClass().getResourceAsStream( "test_Dynamic3_0.drl" ) ) );
        pkg = serialisePackage( builder.getPackage() );
        
        System.out.println("Adding Pkg: org.drools.test with Rule: 'match Person'");
        ruleBase.addPackage( pkg );
        
        session.insert( new Cheese( "stilton", 20 ) );
        session.insert( new Cheese( "brie", 20 ) );
        session.insert( new Person( "bob", 20 ) );
        session.insert( new Person( "bob", 30 ) );
        System.out.println("Firing Rules....");
        session.fireAllRules();
 
        serializedSession = null;
        serializedRulebase = null;
        
        serializedSession = serializeOut( session );
        serializedRulebase = serializeOut( ruleBase );
        
        session.dispose();
        
        System.out.println("3rd Run..");
        
        // now recreate the rulebase, deserialize the session and test it
        ruleBase = (RuleBase) serializeIn( serializedRulebase );
        
        session = ruleBase.newStatefulSession( new ByteArrayInputStream( serializedSession ) );
       
        /*
         * Tried Both:
         *   1. Remove a rule from a pkg.
         *   2. Remove a pkg
         *  But both failed.
         */
        
        System.out.println("Removing Rule: 'like stilton' from Pkg: org.drools.test ");
        ruleBase.removeRule("org.drools.test", "like stilton");
        
       // System.out.println("Removing Pkg: org.drools.newtest ");
       // ruleBase.removePackage("org.drools.newtest");
        
     
      
        session.insert( new Cheese( "stilton", 40 ) );
        session.insert( new Cheese( "brie", 40 ) );
        session.insert( new Person( "bob", 40 ) );
        session.insert( new Person( "bob", 40 ) );

        
        System.out.println("Firing Rules....");
        session.fireAllRules();
 
        serializedSession = null;
        serializedRulebase = null;
        
        serializedSession = serializeOut( session );
        serializedRulebase = serializeOut( ruleBase );
        
        session.dispose();
       
    }
    
    protected Package serialisePackage(Package pkg) {
        try {
        byte[] bytes = serializeOut( pkg );
        return (Package) serializeIn( bytes );
        } catch ( Exception e ) {
            throw new RuntimeException( "trouble serialising package.", e);
        }
    }

    protected Object serializeIn(final byte[] bytes) throws IOException,
                                                    ClassNotFoundException {
        final ObjectInput in = new ObjectInputStream( new ByteArrayInputStream( bytes ) );
        final Object obj = in.readObject();
        in.close();
        return obj;
    }

    protected byte[] serializeOut(final Object obj) throws IOException {
        // Serialize to a byte array
        final ByteArrayOutputStream bos = new ByteArrayOutputStream();
        final ObjectOutput out = new ObjectOutputStream( bos );
        out.writeObject( obj );
        out.close();

        // Get the bytes of the serialized object
        final byte[] bytes = bos.toByteArray();
        return bytes;
    }

}
